"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NitroRpcClient = void 0;
const utils_1 = require("./utils");
const http_1 = require("./transport/http");
const serde_1 = require("./serde");
const constants_1 = require("./constants");
class NitroRpcClient {
    transport;
    // We fetch the address from the RPC server on first use
    myAddress;
    authToken;
    get Notifications() {
        return this.transport.Notifications;
    }
    async CreateVoucher(channelId, amount) {
        const payload = {
            Amount: amount,
            Channel: channelId,
        };
        const request = (0, utils_1.generateRequest)("create_voucher", payload, this.authToken || "");
        const res = await this.transport.sendRequest(request);
        return (0, serde_1.getAndValidateResult)(res, "create_voucher");
    }
    async ReceiveVoucher(voucher) {
        const request = (0, utils_1.generateRequest)("receive_voucher", voucher, this.authToken || "");
        const res = await this.transport.sendRequest(request);
        return (0, serde_1.getAndValidateResult)(res, "receive_voucher");
    }
    async WaitForObjectiveToComplete(objectiveId) {
        const promise = new Promise((resolve) => {
            this.transport.Notifications.on("objective_completed", (payload) => {
                if (payload === objectiveId) {
                    resolve();
                }
            });
        });
        return promise;
    }
    async WaitForLedgerChannelStatus(channelId, status) {
        const promise = new Promise((resolve) => {
            this.transport.Notifications.on("ledger_channel_updated", (payload) => {
                if (payload.ID === channelId) {
                    this.GetLedgerChannel(channelId).then((l) => {
                        if (l.Status == status)
                            resolve();
                    });
                }
            });
        });
        const ledger = await this.GetLedgerChannel(channelId);
        if (ledger.Status == status)
            return;
        return promise;
    }
    async WaitForPaymentChannelStatus(channelId, status) {
        const promise = new Promise((resolve) => {
            this.transport.Notifications.on("payment_channel_updated", (payload) => {
                if (payload.ID === channelId) {
                    this.GetPaymentChannel(channelId).then((l) => {
                        console.log("DEBUG: rpc-client.ts-WaitForPaymentChannelStatus payment channel status after payment_channel_updated notification", l.Status);
                        if (l.Status == status)
                            resolve();
                    });
                }
            });
        });
        const channel = await this.GetPaymentChannel(channelId);
        if (channel.Status == status)
            return;
        return promise;
    }
    onPaymentChannelUpdated(channelId, callback) {
        const wrapperFn = (info) => {
            if (info.ID.toLowerCase() == channelId.toLowerCase()) {
                callback(info);
            }
        };
        this.transport.Notifications.on("payment_channel_updated", wrapperFn);
        return () => {
            this.transport.Notifications.off("payment_channel_updated", wrapperFn);
        };
    }
    async CreateLedgerChannel(counterParty, assetsData, challengeDuration) {
        const payload = {
            CounterParty: counterParty,
            ChallengeDuration: challengeDuration,
            Outcome: (0, utils_1.createOutcome)(await this.GetAddress(), counterParty, assetsData),
            AppDefinition: constants_1.ZERO_ETHEREUM_ADDRESS,
            AppData: "0x00",
            Nonce: Date.now(),
        };
        return this.sendRequest("create_ledger_channel", payload);
    }
    async RetryObjectiveTx(objectiveId) {
        return this.sendRequest("retry_objective_tx", { ObjectiveId: objectiveId });
    }
    async RetryTx(txHash) {
        return this.sendRequest("retry_tx", { TxHash: txHash });
    }
    async CreateSwapChannel(counterParty, intermediaries, assetsData) {
        const payload = {
            CounterParty: counterParty,
            Intermediaries: intermediaries,
            ChallengeDuration: 0,
            Outcome: (0, utils_1.createOutcome)(await this.GetAddress(), counterParty, assetsData),
            AppDefinition: constants_1.ZERO_ETHEREUM_ADDRESS,
            Nonce: Date.now(),
        };
        return this.sendRequest("create_swap_channel", payload);
    }
    async SwapAssets(channelId, swapAssetsData) {
        const payload = {
            SwapAssetsData: swapAssetsData,
            Channel: channelId,
        };
        const request = (0, utils_1.generateRequest)("swap_initiate", payload, this.authToken || "");
        const res = await this.transport.sendRequest(request);
        return (0, serde_1.getAndValidateResult)(res, "swap_initiate");
    }
    async GetPendingSwap(channelId) {
        const payload = {
            Id: channelId,
        };
        return this.sendRequest("get_pending_swap", payload);
    }
    async GetRecentSwaps(channelId) {
        const payload = {
            Id: channelId,
        };
        return this.sendRequest("get_recent_swaps", payload);
    }
    async CreatePaymentChannel(counterParty, intermediaries, amount) {
        const asset = `0x${"00".repeat(20)}`;
        const payload = {
            CounterParty: counterParty,
            Intermediaries: intermediaries,
            ChallengeDuration: 0,
            Outcome: (0, utils_1.createOutcome)(await this.GetAddress(), counterParty, [
                {
                    assetAddress: asset,
                    alphaAmount: amount,
                    // As payment channel is simplex, only alpha node can pay beta node and not vice-versa hence beta node's allocation amount is 0
                    betaAmount: 0,
                },
            ]),
            AppDefinition: asset,
            Nonce: Date.now(),
        };
        return this.sendRequest("create_payment_channel", payload);
    }
    async Pay(channelId, amount) {
        const payload = {
            Amount: amount,
            Channel: channelId,
        };
        const request = (0, utils_1.generateRequest)("pay", payload, this.authToken || "");
        const res = await this.transport.sendRequest(request);
        return (0, serde_1.getAndValidateResult)(res, "pay");
    }
    async GetVoucher(channelId) {
        const payload = {
            Id: channelId,
        };
        return this.sendRequest("get_voucher", payload);
    }
    async CloseLedgerChannel(channelId, isChallenge) {
        const payload = {
            ChannelId: channelId,
            IsChallenge: isChallenge,
        };
        return this.sendRequest("close_ledger_channel", payload);
    }
    async CloseBridgeChannel(channelId) {
        const payload = {
            ChannelId: channelId,
        };
        return this.sendRequest("close_bridge_channel", payload);
    }
    async MirrorBridgedDefund(channelId, stringifiedL2SignedState, isChallenge) {
        const payload = {
            ChannelId: channelId,
            IsChallenge: isChallenge,
            StringifiedL2SignedState: stringifiedL2SignedState,
        };
        return this.sendRequest("mirror_bridged_defund", payload);
    }
    async CounterChallenge(channelId, action, signedState) {
        const payload = {
            ChannelId: channelId,
            Action: action,
            StringifiedL2SignedState: signedState,
        };
        return this.sendRequest("counter_challenge", payload);
    }
    async ConfirmSwap(swapId, action) {
        const payload = {
            SwapId: swapId,
            Action: action,
        };
        return this.sendRequest("confirm_swap", payload);
    }
    async ClosePaymentChannel(channelId) {
        const payload = { ChannelId: channelId };
        return this.sendRequest("close_payment_channel", payload);
    }
    async CloseSwapChannel(channelId) {
        const payload = { ChannelId: channelId };
        return this.sendRequest("close_swap_channel", payload);
    }
    async GetVersion() {
        return this.sendRequest("version", {});
    }
    async GetNodeInfo() {
        return this.sendRequest("get_node_info", {});
    }
    async GetAddress() {
        if (this.myAddress) {
            return this.myAddress;
        }
        this.myAddress = await this.sendRequest("get_address", {});
        return this.myAddress;
    }
    async GetLedgerChannel(channelId) {
        return this.sendRequest("get_ledger_channel", { Id: channelId });
    }
    async GetAllLedgerChannels() {
        return this.sendRequest("get_all_ledger_channels", {});
    }
    async GetAllL2Channels() {
        return this.sendRequest("get_all_l2_channels", {});
    }
    async GetSignedState(channelId) {
        return this.sendRequest("get_signed_state", { Id: channelId });
    }
    async GetPaymentChannel(channelId) {
        return this.sendRequest("get_payment_channel", { Id: channelId });
    }
    async GetSwapChannel(channelId) {
        return this.sendRequest("get_swap_channel", { Id: channelId });
    }
    async GetPaymentChannelsByLedger(ledgerId) {
        return this.sendRequest("get_payment_channels_by_ledger", {
            LedgerId: ledgerId,
        });
    }
    async GetSwapChannelsByLedger(ledgerId) {
        return this.sendRequest("get_swap_channels_by_ledger", {
            LedgerId: ledgerId,
        });
    }
    async GetObjective(objectiveId, l2) {
        return this.sendRequest("get_objective", {
            ObjectiveId: objectiveId,
            L2: l2,
        });
    }
    async GetL2ObjectiveFromL1(l1ObjectiveId) {
        return this.sendRequest("get_l2_objective_from_l1", {
            L1ObjectiveId: l1ObjectiveId,
        });
    }
    async GetPendingBridgeTxs(channelId) {
        return this.sendRequest("get_pending_bridge_txs", {
            ChannelId: channelId,
        });
    }
    async getAuthToken() {
        return this.sendRequest("get_auth_token", {});
    }
    async sendRequest(method, payload) {
        const request = (0, utils_1.generateRequest)(method, payload, this.authToken || "");
        const res = await this.transport.sendRequest(request);
        return (0, serde_1.getAndValidateResult)(res, method);
    }
    async Close() {
        return this.transport.Close();
    }
    constructor(transport) {
        this.transport = transport;
    }
    /**
     * Creates an RPC client that uses HTTP/WS as the transport.
     *
     * @param url - The URL of the HTTP/WS server
     * @returns A NitroRpcClient that uses WS as the transport
     */
    static async CreateHttpNitroClient(url, isSecure) {
        const transport = await http_1.HttpTransport.createTransport(url, isSecure);
        const rpcClient = new NitroRpcClient(transport);
        rpcClient.authToken = await rpcClient.getAuthToken();
        return rpcClient;
    }
}
exports.NitroRpcClient = NitroRpcClient;
