"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertToSwapChannelsInfoType = exports.convertToSwapChannelInfoType = exports.getAndValidateNotification = exports.getAndValidateResult = void 0;
const jtd_1 = __importDefault(require("ajv/dist/jtd"));
const json_bigint_1 = __importDefault(require("json-bigint"));
const types_1 = require("./types");
const ajv = new jtd_1.default();
const jsonRpcSchema = {
    properties: {
        jsonrpc: { type: "string" },
        id: { type: "uint32" },
    },
    optionalProperties: {
        result: {
            nullable: true,
        },
        error: {
            properties: {
                code: { type: "int32" },
                message: { type: "string" },
            },
            additionalProperties: true,
            nullable: true,
        },
    },
};
const objectiveSchema = {
    properties: {
        Id: { type: "string" },
        ChannelId: { type: "string" },
    },
};
const stringSchema = { type: "string" };
const counterChallengeSchema = {
    properties: {
        ChannelId: { type: "string" },
        Action: { type: "int32" },
        StringifiedL2SignedState: { type: "string" },
    },
};
const confirmSwapSchema = {
    properties: {
        SwapId: { type: "string" },
        Action: { type: "int32" },
    },
};
const ledgerChannelSchema = {
    properties: {
        ID: { type: "string" },
        Status: { type: "string" },
        ChannelMode: { type: "int32" },
        Balances: {
            elements: {
                properties: {
                    AssetAddress: { type: "string" },
                    Them: { type: "string" },
                    Me: { type: "string" },
                    MyBalance: { type: "string" },
                    TheirBalance: { type: "string" },
                },
            },
        },
    },
};
const paymentChannelSchema = {
    properties: {
        ID: { type: "string" },
        Status: { type: "string" },
        Balance: {
            properties: {
                AssetAddress: { type: "string" },
                Payee: { type: "string" },
                Payer: { type: "string" },
                PaidSoFar: { type: "string" },
                RemainingFunds: { type: "string" },
            },
        },
    },
};
const swapChannelSchema = {
    properties: {
        ID: { type: "string" },
        Status: { type: "string" },
        Balances: {
            // TODO: Check array inside properties
            elements: {
                properties: {
                    AssetAddress: { type: "string" },
                    NodeA: { type: "string" },
                    NodeB: { type: "string" },
                    AmountNodeA: { type: "string" },
                    AmountNodeB: { type: "string" },
                },
            },
        },
    },
};
const ledgerChannelsSchema = {
    elements: {
        ...ledgerChannelSchema,
    },
};
const paymentChannelsSchema = {
    elements: {
        ...paymentChannelSchema,
    },
};
const paymentSchema = {
    properties: {
        Amount: { type: "uint32" },
        Channel: { type: "string" },
    },
};
const swapSchema = {
    properties: {
        SwapAssetsData: {
            properties: {
                TokenIn: { type: "string" },
                TokenOut: { type: "string" },
                AmountIn: { type: "uint32" },
                AmountOut: { type: "uint32" },
            },
        },
        Channel: { type: "string" },
    },
};
const voucherSchema = {
    properties: {
        ChannelId: { type: "string" },
        Amount: { type: "uint32", nullable: true },
        Signature: {
            type: "string",
        },
    },
};
const receiveVoucherSchema = {
    properties: {
        Total: { type: "string" },
        Delta: { type: "string" },
    },
};
const getNodeInfoSchema = {
    properties: {
        SCAddress: { type: "string" },
        MessageServicePeerId: { type: "string" },
    },
};
/**
 * Validates that the response is a valid JSON RPC response with a valid result
 * @param response - JSON RPC response
 * @param method - JSON RPC method
 * @returns The validated result of the JSON RPC response
 */
function getAndValidateResult(response, method) {
    const { result, error } = getJsonRpcResult(response);
    if (error) {
        throw new Error("jsonrpc response: " + error.message);
    }
    switch (method) {
        case "create_ledger_channel":
        case "create_payment_channel":
        case "create_swap_channel":
            return validateAndConvertResult(objectiveSchema, result, (result) => result);
        case "get_pending_bridge_txs":
        case "get_l2_objective_from_l1":
        case "retry_objective_tx":
        case "retry_tx":
        case "get_objective":
        case "get_auth_token":
        case "close_ledger_channel":
        case "close_bridge_channel":
        case "mirror_bridged_defund":
        case "version":
        case "get_address":
        case "get_signed_state":
        case "close_payment_channel":
        case "close_swap_channel":
        case "get_pending_swap":
        case "get_recent_swaps":
            return validateAndConvertResult(stringSchema, result, (result) => result);
        case "get_swap_channel":
            return validateAndConvertResult(stringSchema, result, exports.convertToSwapChannelInfoType);
        case "get_swap_channels_by_ledger":
            return validateAndConvertResult(stringSchema, result, exports.convertToSwapChannelsInfoType);
        case "get_ledger_channel":
            return validateAndConvertResult(ledgerChannelSchema, result, convertToInternalLedgerChannelType);
        case "counter_challenge":
            return validateAndConvertResult(counterChallengeSchema, result, convertToCounterChallengeResultType);
        case "confirm_swap":
            return validateAndConvertResult(confirmSwapSchema, result, convertToConfirmSwapResultType);
        case "get_all_ledger_channels":
            return validateAndConvertResult(ledgerChannelsSchema, result, convertToInternalLedgerChannelsType);
        case "get_all_l2_channels":
            return validateAndConvertResult(ledgerChannelsSchema, result, convertToInternalLedgerChannelsType);
        case "get_payment_channel":
            return validateAndConvertResult(paymentChannelSchema, result, convertToInternalPaymentChannelType);
        case "get_payment_channels_by_ledger":
            return validateAndConvertResult(paymentChannelsSchema, result, convertToInternalPaymentChannelsType);
        case "pay":
            return validateAndConvertResult(paymentSchema, result, (result) => result);
        case "swap_initiate":
            return validateAndConvertResult(swapSchema, result, (result) => result);
        case "receive_voucher":
            return validateAndConvertResult(receiveVoucherSchema, result, (result) => ({
                Total: BigInt(result.Total),
                Delta: BigInt(result.Delta),
            }));
        case "get_voucher":
        case "create_voucher":
            return validateAndConvertResult(voucherSchema, result, (result) => ({
                Amount: result.Amount ?? 0,
                ChannelId: result.ChannelId,
                Signature: result.Signature,
            }));
        case "get_node_info":
            return validateAndConvertResult(getNodeInfoSchema, result, (result) => result);
        default:
            throw new Error(`Unknown method: ${method}`);
    }
}
exports.getAndValidateResult = getAndValidateResult;
function getAndValidateNotification(data, method) {
    switch (method) {
        case "payment_channel_updated":
            return convertToInternalPaymentChannelType(data);
        case "ledger_channel_updated":
            return convertToInternalLedgerChannelType(data);
        case "objective_completed":
            return data;
        default:
            throw new Error(`Unknown method: ${method}`);
    }
}
exports.getAndValidateNotification = getAndValidateNotification;
/**
 * Validates that the response is a valid JSON RPC response and pulls out the result
 * @param response - JSON RPC response
 * @returns The result of the response
 */
function getJsonRpcResult(response) {
    const validate = ajv.compile(jsonRpcSchema);
    if (validate(response)) {
        return response;
    }
    throw new Error(`Invalid json rpc response: ${JSON.stringify(validate.errors)}. The response is ${JSON.stringify(response)}`);
}
/**
 * validateAndConvertResult validates that the result object conforms to the schema and converts it to the internal type
 *
 * @param schema - JSON Type Definition
 * @param result - Object to validate
 * @param converstionFn - Function to convert the valiated object to internal type
 * @returns A validated object of internal type
 */
function validateAndConvertResult(schema, result, converstionFn) {
    const validate = ajv.compile(schema);
    if (validate(result)) {
        return converstionFn(result);
    }
    throw new Error(`Error parsing json rpc result: ${JSON.stringify(validate.errors)}. The result is ${JSON.stringify(result)}`);
}
function convertToInternalLedgerChannelType(result) {
    // todo: validate channel status
    return {
        ...result,
        Status: result.Status,
        ChannelMode: types_1.ChannelMode[result.ChannelMode],
        Balances: result.Balances.map((balance) => ({
            ...balance,
            MyBalance: BigInt(balance.MyBalance ?? 0),
            TheirBalance: BigInt(balance.TheirBalance ?? 0),
        })),
    };
}
const convertToSwapChannelInfoType = (result) => {
    const swapChannelInfo = JSON.parse(result);
    const modifiedSwapChannelInfo = {
        ...swapChannelInfo,
        Balances: swapChannelInfo.Balances.map((balance) => ({
            ...balance,
            MyBalance: BigInt(balance.MyBalance),
            TheirBalance: BigInt(balance.TheirBalance),
        })),
    };
    return modifiedSwapChannelInfo;
};
exports.convertToSwapChannelInfoType = convertToSwapChannelInfoType;
const convertToSwapChannelsInfoType = (result) => {
    const swapChannelsInfo = JSON.parse(result);
    const modifiedSwapChannelsInfo = swapChannelsInfo.map((sc) => {
        return {
            ...sc,
            Balances: sc.Balances.map((balance) => ({
                ...balance,
                MyBalance: BigInt(balance.MyBalance),
                TheirBalance: BigInt(balance.TheirBalance),
            })),
        };
    });
    return json_bigint_1.default.stringify(modifiedSwapChannelsInfo, null, 2);
};
exports.convertToSwapChannelsInfoType = convertToSwapChannelsInfoType;
function convertToCounterChallengeResultType(result) {
    return {
        ChannelId: result.ChannelId,
        Action: types_1.CounterChallengeAction[result.Action],
    };
}
function convertToConfirmSwapResultType(result) {
    return {
        SwapId: result.SwapId,
        Action: types_1.ConfirmSwapAction[result.Action],
    };
}
function convertToInternalLedgerChannelsType(result) {
    return result.map((lc) => convertToInternalLedgerChannelType(lc));
}
function convertToInternalPaymentChannelType(result) {
    // todo: validate channel status
    return {
        ...result,
        Status: result.Status,
        Balance: {
            ...result.Balance,
            PaidSoFar: BigInt(result.Balance.PaidSoFar ?? 0),
            RemainingFunds: BigInt(result.Balance.RemainingFunds ?? 0),
        },
    };
}
function convertToInternalPaymentChannelsType(result) {
    return result.map((pc) => convertToInternalPaymentChannelType(pc));
}
